#!/usr/bin/perl -w
# relay.pl: A script to relay commands to IRC from a webpage. 1000 lines.
# Copyright Ian Clark. Do not redistribute. Valid download location is:
#   <http://dev-com.quakenet.org/> If you didn't receive a .sig file this may
#   be trojaned. No support at all, ever.
$version = "Relay/0.5.0";
###############################################################################
# External functions/classes
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
use IO::Socket;   # For network stuff
use IO::Select;   # For select() 
use Fcntl;        # To make the sockets non-blocking
use URI::Escape;  # To convert URL encoded stuff to real text
use MIME::Base64; # For http-basic-auth. 

###############################################################################
# Initialisation stuff. Messy
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Pre-declare some items to keep warnings quiet ;)
($opass, $listenport, $quser, $qpass, $ouser, $opass, $port, $logfile, $userfile,$stone) = "";
require("relay.conf"); # Load config file
system($stone); # This does the SSL/TLS stuff. Logs to stone.log, usually.
open(LOG,">>$logfile");
plog("Logfile opened\n");
plog("$version\n");
open(USERS,$userfile);
@users = <USERS>; # Sluuurp.
close(USERS);

$SIG{INT} = sub { # Signal handler. Shuts down cleanly and saves stuff if it can
  $select->remove($irc);
  $irc->send("QUIT :Received SIGINT\n",0);
  $irc->close;
  $select->remove($listen);
  $listen->close;
  &savequeue();
  plog("SIGINT Received\n");
  die("SIGINT Received.\n");
};

# Global vars
@queue=(); # Array of references to hashes, like the one below
$current={conn=> "", command=>"", user=>"", pass=>"",
          chan=> "", page=>   "", mail=>"", temp=>""};
$ready=1;  # Flag to show queue status. If 0 something is happening.
$working=0;# Flag to disable system when not connected, or Q hiding somewhere

# CHECK status info. This is a pain.
$check = 0;
$chandata = { chan  => "", ctime => "", topic => "",
	      modes => "", users => "", stats => ""};

@reqqueue = (); # Queue for requests
$requests = 0;  # Size of the queue
&loadqueue();

$loop=0; # Stupid loop to handle when the listen address temporarily won't bind
while(!($listen = IO::Socket::INET->new(Listen => 1, 
                                        LocalPort => $listenport,
					LocalAddr => "localhost"))) {
  sleep 20;
  $loop++;
  if($loop == 12) {
    die("Couldn't listen after 4mins\n");
  }
}
undef($loop);
fcntl($listen,F_SETFL,O_NONBLOCK); # Don't want this blocking on us..

# Connect to IRC
$irc = IO::Socket::INET->new("$server:$port") or 
  die("Connection problem: $!\n"); 
fcntl($irc,F_SETFL,O_NONBLOCK); # or this.

&ircsend("USER relay null null :$version\n");
&ircsend("NICK $botnick\n");

# Prepare select..
$select = IO::Select->new($irc,$listen,\*STDIN);

###############################################################################
# Main loop. Checks for activity on all current connections, parses queue.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
while(@ready = $select->can_read(240)) {
  foreach my $fh (@ready) {
    if($fh == $listen) {
      $new = $listen->accept;
      $select->add($new);
    } elsif($fh == $irc) {
      &ircparse();
    } elsif($fh == \*STDIN) {
      my $foo = <STDIN>; # Yes, I am silly
      if($foo =~ /^q/i) {
        my @handles = $select->handles;
	&ircsend("QUIT :Quitting, Console command\n");
        $select->remove(@handles);
	foreach $fh (@handles) {
	  if($fh != \*STDIN) {
            $fh->close;
	  }
	}
	&savequeue();
	plog("Quit. Console command\n");
	exit(0);
      }
    } else {
      &httpparse($fh);
    }  
  }
  if($ready == 1) { # Possible prob: this will only be parsed when there's 
                    # activity on irc/http. Should be OK, because the queue
		    # is only added to when this occurs.
    if($current = shift(@queue)) { 
      # process instructions here
      if($$current{'command'} =~ /checkauth/) { &checkauth(); };
      if($$current{'command'} =~ /checkchan/) { &checkchan(); };
      if($$current{'command'} =~ /addchan/)   { &addchan();   };
      if($$current{'command'} =~ /operpass/)  { 
        $ready=0;
	&ircsend("WHOIS $$current{'user'}\n");
      }
    }
  }
}
&savequeue();
die("Timeout.");

###############################################################################
# Parsing subroutines. These tend to give the work to other subs.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# The first line of HTTP traffic is parsed here. The rest is left to the HTTP
#   subroutines.
sub httpparse {
  my $conn = shift;
  my $buffer;
  my $peer = $conn->peerhost . ":" . $conn->peerport;
  if(!($buffer = $conn->getline)) { # Only want one line here
    $select->remove($conn);
    $conn->close;
    return;
  }
  print("\033[31m <IN>\033[0m [$peer] $buffer"); 
  plog("[$peer] $buffer");
  if($working==0)                            { &show_notready($conn); return; };
  if($buffer =~ /^GET \/ HTTP/)              { &show_welcome($conn);  return; };
  if($buffer =~ /^POST \/login HTTP/)        { &login_info($conn);    return; };
  if($buffer =~ /^GET \/queue HTTP/)         { &show_queue($conn);    return; };
  if($buffer =~ /^GET \/accept\/(\S+) HTTP/) { &accept($conn,$1);     return; };
  if($buffer =~ /^GET \/reject\/(\S+) HTTP/) { &reject($conn,$1);     return; };

  print("\033[31;1m*ERR*\033[0m 404 Not Found\n");
  plog("404 Not Found\n");
  &connsend($conn,<<"EOF");
HTTP/1.1 404 Not Found
Server: $version
Connection: close
Content-Type: text/html

<html>
 <head>
  <title>404 Not Found</title>
 </head>
 <body>
  <h1>404 Not Found</h1>
  <p>The requested URL was not found</p>
 </body>
</html>
EOF
  $select->remove($conn);
  $conn->close;
}

# IRC stuff parsed and acted on here.
sub ircparse {
  while(my $buffer = $irc->getline) {
    print("\033[32m<IRC>\033[0m ");
    
    $buffer =~ s/\002//g; # Strips bolds (durzel must die!)
    print($buffer);

    if($buffer =~ /:(\S+)!\S+ PRIVMSG $botnick :PASS (\S+)/i) {
      push(@queue,{user=>$1,pass=>$2,command=>"operpass"});
    }

    # Is an ircop
    if($buffer =~ /$server 313/i) {
      if($$current{'command'} =~ /operpass/) {
        push(@users,"$$current{'user'}:$$current{'pass'}");
        open(USERS,">>$userfile");
        print(USERS "$$current{'user'}:$$current{'pass'}\n");
        close(USERS);
        &ircsend("PRIVMSG $$current{'user'} :OK. Password set.\n");
        $ready=1;
      }
    }

    # End of WHOIS
    if($buffer =~ /$server 318/i) {
      if($$current{'command'} =~ /operpass/) {
        $ready=1;
      }
    }

    # Initial PING on ircu derived servers.
    if($buffer =~ /^PING :(\d+)/i) { 
      &ircsend("PONG $1\n"); 
    }

    # End of motd stuff.
    if($buffer =~ /^:$server 376/i) {
      &ircsend("OPER $ouser $opass\n");
      &ircsend("MODE $botnick +dkX\n");
      &ircsend("JOIN #twilightzone\n");
      &ircsend("ISON Q\n");
    }

    # Normal server PINGs. Also everything we need to do semi-frequently.
    if($buffer =~ /^PING :$server/i) { 
      &ircsend("PONG $botnick\n"); 
      &ircsend("ISON Q\n");
      if($requests > 0) {
        &ircsend("PRIVMSG #twilightzone :$requests item(s) in queue: $myurl\n");
	&savequeue();
      }
      $working = 0;
    }

    # No such nick
    if($buffer =~ /:$server 401 $botnick Q/i) {
        &connsend($$current{'conn'},<<"EOF");
 Failed: Q: no such nick</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
        $ready=1;
	$working=0;
	shift(@queue);
        $select->remove($$current{'conn'});
        $$current{'conn'}->close;
    }

    if($buffer =~ /:$server 403 $botnick/i) {
        &connsend($$current{'conn'},<<"EOF");
 Failed: no such channel</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
        $ready=1;
	$select->remove($$current{'conn'});
	$$current{'conn'}->close;
    }

    # Grabbing own ip from join info
    if($buffer =~ /$botnick\S+\@(\S+) JOIN/i) {
      $myurl = "https://$1:8000/queue";
      &ircsend("MODE $botnick -X\n");
    }

    # Q verification part 1
    if($buffer =~ /:$server 303 $botnick :Q/i) {
      &ircsend("USERHOST Q\n");
    }

    # Q verification part 2
    if($buffer =~ /:$server 302 $botnick :Q\*=.$qhost/i) {
      $working = 1;
      &ircsend("PRIVMSG Q :AUTH $quser $qpass\n");
    }
  
    # Q interaction
    if($buffer =~ /:Q!$qhost NOTICE $botnick :(.+)$/i) { 
      &qparse($1); 
    }

    # CHECK parsing
    if($buffer =~ /:$server 286 $botnick :Information for channel (\S+)/i) {
      my $chan=$1;
      $chan =~ s/\'//g; # To handle different versions of qnet patch.
      $check = 1;       #   Damn you Durzel.
      $chandata = { 
	      chan  => "",
              ctime => "",
	      topic => "",
	      modes => "",
	      users => "",
	      stats => ""};
      $$chandata{'chan'}=$chan;
    }
    if($check) {
      if($buffer =~ /:$server 257 $botnick :.+Creation time:: (.+)$/i) {
        $$chandata{'ctime'}=$1;
      }
      if($buffer =~ /:$server 257 $botnick :.+Topic:: (.+)$/i) {
        $$chandata{'topic'}=$1;
      }
      if($buffer =~ /:$server 257 $botnick :.*Channel mode.s.:: (.+)$/i) {
	$$chandata{'modes'}=($1);
	chomp($$chandata{'modes'});
      }
      if($buffer =~ /:$server 287 $botnick :\s+.\s+(\S+) \((\S+)\)\s+(.+)$/i) {
	my $nick = $1;
	my $host = $2;
	my $ircname = $3;
	$ircname =~ s/\r//; # Remove CR
	$$chandata{'users'}=$$chandata{'users'} . "$nick!$host$ircname<br />\n";
      }
      if($buffer =~ /:$server 257 $botnick :Total users:: (\d+) \S(\d+) ops, \d+ voiced, (\d+) clones\S/i) {
        if($2 < 4) { 
          &connsend($$current{'conn'},<<"EOF");
 Failed: You have too few ops on your channel. Q cannot be added automatically</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
          $select->remove($$current{'conn'});
	  $$current{'conn'}->close;
	  $check=0;
	  $ready=1;
	  next;
        }
	if($3 > 2) { # Allow 2 clones, in case of someone reconnecting.
	  &connsend($$current{'conn'},<<"EOF");
 Failed: You have clones on your channel. Q cannot be added automatically</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
          $select->remove($$current{'conn'});
	  $$current{'conn'}->close;
	  $check=0;
	  $ready=1;
	  next;
	}
        $$chandata{'stats'}="U: $1 O: $2 C: $3";
      }
      if($buffer =~ /:$server 257 $botnick :Bans/i) {
        &addtoqueue();
      }
    }
  }
}

# Responses from Q parsed here
sub qparse {
  my $response = shift;
  # Not authed...
  if($response =~ /You don.t have access/i) {
    $working = 0;
    &ircsend("ISON Q\n"); # Trigger Q verification/auth process
    unshift(@queue,$current); # put thing back in queue for now
    $ready=1;
  }

  # Password request OK
  if($response =~ /Password is currently: (\S+)/i) {
    my $realpass = $1;
    if($$current{'command'} =~ /checkauth/) { 
      if($$current{'pass'} =~ /^$realpass$/) { # need this to match exactly
        # Password was correct
        &connsend($$current{'conn'},<<"EOF");
 OK.</li>
   <li>Checking channel...
EOF
        $ready=1;
      } else {
        # Password incorrect
        shift(@queue); # remove checkchan request
        &connsend($$current{'conn'},<<"EOF");
 Failed. Incorrect password. Try again</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
        $ready=1;
        $select->remove($$current{'conn'});
        $$current{'conn'}->close;
      }
    } else {
      # Got a password response we didn't ask for.
      # Of course, this will never happen *cough*
      print("\033[31;1m*ERR*\033[0m Unrequested password!\n");
    }   
  }

  # Password request failed
  if($response =~ /Couldn.t find that user in/i) {
    shift(@queue);
    &connsend($$current{'conn'},<<"EOF");
 Failed: No such user. Try again</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
    $ready=1;
    $select->remove($$current{'conn'});
    $$current{'conn'}->close;
  }
  if($response =~ /User has authlevel/i) {
    shift(@queue);
    &connsend($$current{'conn'},<<"EOF");
 Failed: That username has auth level > 100 and cannot use this system</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
    $ready=1;
    $select->remove($$current{'conn'});
    $$current{'conn'}->close;
  }

  # Finding user's current nickname (part of checkchan)
  if($response =~ /(\S+) is authed as $$current{'user'}/i) {
    $$current{'user'} = $$current{'user'} . "/$1";
#    &ircsend("PRIVMSG $$current{'chan'} :$$current{'user'} has requested Q for this channel\n");
    &ircsend("CHECK $$current{'chan'}\n");
  }

  # multiple auths
  if($response =~ /There are more users authed with that id./i) {
    &connsend($$current{'conn'},<<"EOF");
 Failed: More than one user is authed with that username</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
    $ready=1;
    $select->remove($$current{'conn'});
    $$current{'conn'}->close;
  }

  # end of Q WHOIS. If nick not set, go moo
  if($response =~ /(End of list|User not known in any channels)/i) {
    if($$current{'user'} !~ /\//) {
      &connsend($$current{'conn'},<<"EOF");
 Failed: You do not seem to be currently authed on Q</li>
  </ul>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
      $ready=1;
      $select->remove($$current{'conn'});
      $$current{'conn'}->close;
    }
  }
 
  # Addchan successful
  if($response =~ /done./i) {
    &remrequest($$current{'chan'});
    &connsend($$current{'conn'},<<"EOF");
HTTP/1.1 200 OK
Server: $version
Connection: keep-alive
Content-type: text/html

<html>
 <head>
  <title>$$current{'chan'} accepted</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
  <p>Channel addition complete.</p>
  <p>[<a href="/queue">Back</a>]</p>
 </body>
</html>
EOF
    $select->remove($$current{'conn'});
    $$current{'conn'}->close;
    $ready=1;
  }

  # Addchan failed: TODO
  if($response =~ /Mooo(\S+)/i) {
    my $url = uri_escape($$current{'chan'});
    &connsend($$current{'conn'},<<"EOF");
HTTP/1.1 200
Server: $version
Connection: keep-alive
Content-type: text/html

<html>
 <head>
  <title>Error adding $$current{'chan'}</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
  <p>When trying to add channel, Q said $1</p>
  <p><a href=\"/accept/$url\">Retry</a>/<a href=\"/reject/$url\">Give up</a></p>
 </body>
</html>
EOF
    $select->remove($$current{'conn'});
    $$current{'conn'}->close;
    $ready=1;
  }
}

###############################################################################
# Queue Handlers: called by the queue parser, generally trigger events in irc
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub checkchan {
  print("\033[34m  ***\033[0m CheckChan\n");
  $ready=0;
  &ircsend("PRIVMSG Q :WHOIS #$$current{'user'}\n");
}

sub checkauth {
  $ready=0;
  print("\033[34m  ***\033[0m CheckAuth\n");
  &ircsend("PRIVMSG Q :GETPASS #$$current{'user'}\n");
}

sub addchan {
  print("\033[34m  ***\033[0m AddChan\n");
  $ready=0;
  my($user) = $$current{'user'};
  $user =~ s/\/\S+//;
  &ircsend("PRIVMSG Q :ADDCHAN $$current{'chan'} + #$user $$current{'type'}\n");
}

###############################################################################
# URL Handlers: called from the HTTP parser, Most finish without triggering
#               anything else.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# If Q is AWOL this is shown.
sub show_notready {
  my $conn = shift;
  &connsend($conn,<<"EOF");
HTTP/1.1 200 OK
Server: $version
Connection: keep-alive
Content-Type: text/html

<html>
 <head>
  <title>Q Bot Request System</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
  <h1>System not currently working</h1>
  <p>Please try again later</p>
  <!-- BattleCow says hi -->
 </body>
</html>
EOF
  $select->remove($conn);
  $conn->close;
}

# Shows the initial page
sub show_welcome {
  my $conn = shift;
  &connsend($conn,<<"EOF");
HTTP/1.1 200 OK
Server: $version
Connection: keep-alive
Content-Type: text/html

<html>
 <head>
  <title>Q Bot Request System</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
  <h1>Enter your details</h1>
  <table style="border: 1px ; border-color: white"><tr><td>
   <form action="/login" method="post">
    <div align="right">
     <label style="font-weight: bold;color: white">Username:<input name="username" /></label><br />
     <label style="font-weight: bold;color: white">Password:<input type="password" name="password" /></label><br />
     <label style="font-weight: bold;color: white">Channel:<input name="channel" /></label><br />
     <lable style="font-weight: bold;color: white">
      <select name="type" width="40">
       <option>Clan</option>
       <option>Private</option>
       <option>Special</option>
       <option>Gamesite</option>
       <option>Game</option>
      </select><br />
     </label>
     <label style="font-weight: bold;color: white">Webpage:<input name="webpage" /></label><br />
     <label style="color: white">Email:<input name="email" /></span></label><br />
     <input type="submit" /><input type="reset" />
    </div>
   </form>
  </td></tr></table>
  <h1>Notes:</h1>
  <ul>
   <li>If you don't have a Q username and password, <code>/msg Q hello</code> on IRC</li>
   <li>You must be authed and opped in the channel, with 3 other clan members
    also opped. (No Clones)</li>
   <li>Channel names must include the "#". Q cannot reserve "&" channels.</li>
   <li>Items in bold are required</li>
   <li>For email confirmation, enter your address in the Email field</li>
  </ul>
  <!-- BattleCow says hi -->
 </body>
</html>
EOF
  $select->remove($conn);
  $conn->close; 
}

# Reads in the details of a Q request. Does initial checks.
sub login_info {
  my $conn = shift;
  &connsend($conn, "HTTP/1.1 200 OK\n");
  my ($user,$pass,$chan,$page,$mail,$contentlength);
  LINE: while(my $line = $conn->getline) {
    if($line !~ /^\S+ .+$/) { last LINE; };
    if($line =~ /^Content-Length: (\d+)/i) { 
      $contentlength = $1; 
    };
  }
  
  # I hate trusting external data. Will have to make this safe.
  if($conn->read($line,$contentlength) < $contentlength) {
    print("\033[31;1m*ERR*\033[0m Content underflow. Dropping connection\n"); 
    plog("Content underflow. Connection was dropped\n");
    $select->remove($conn);
    $conn->close;
  }
  if($line =~ /^username=(\S+)\&password=(\S+)\&channel=(\S+)\&type=(\S+)\&webpage=(\S+)\&email=(\S*)/i) {
    $user=uri_unescape($1);
    $pass=uri_unescape($2);
    $chan=uri_unescape($3);
    $type=uri_unescape($4);
    $page=uri_unescape($5);
    $mail=uri_unescape($6);
    print("\033[36m<REQ>\033[0m user: $user pass: $pass chan: $chan type: $type page: $page mail: $mail\n");
    plog("Request: $user, $pass, $chan, $type, $page, $mail\n");
    &connsend($conn,<<"EOF");
Server: $version
Connection: keep-alive
Content-Type: text/html

<html>
 <head>
  <title>Info Received</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
  <h1>Submitted data</h1>
  <p>Username: $user, Channel: $chan, Type: $type, Webpage: <a href="$page">$page</a>, email: $mail</p>
EOF
    foreach my $temp (@reqqueue) {
      if($$temp{'chan'} =~ /^\Q$chan\E$/) { # Grr. Why do channel names need a #
        &connsend($conn,<<"EOF");
  <p>That channel is already queued</p>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
        $select->remove($conn);
        $conn->close;
        return;
      }
    }
  }
  &connsend($conn,<<"EOF");
  <ul>
   <li>Checking password...
EOF
  $temp = {conn=>$conn,command=>"checkauth",user=>$user,pass=>$pass};
  push(@queue,$temp);
  $temp = {conn=>$conn,command=>"checkchan",user=>$user,chan=>$chan,type=>$type,page=>$page,mail=>$mail};
  push(@queue,$temp);
}

# Requires http-basic-auth before sending queue info
sub show_queue {
  my $conn = shift;
  my $oper = &requireauth($conn);
  if($oper =~ /FAILED/) { return };
 
  print("\033[31mAUTH>\033[0m $oper -^\n");
  plog("AUTH $oper -^\n");
  &connsend($conn,<<"EOF");
HTTP/1.1 200 OK
Server: $version
Www-Authenticate: Basic realm="qweb"
Connection: keep-alive
Content-Type: text/html

<html>
 <head>
  <title>Channels in Request Queue</title>
  <style>
   body { background: black ; color: white }
   table { background: grey90 ; border: 2px ; border-color: white}
   th { background: white ; color: black }
  </style>
 </head>
 <body>
  <table>
   <th><td>Username</td><td>Type</td><td>URL</td><td>CreationTime</td><td>Stats</td><td>Modes</td><td>Topic</td><td>Users</td></th>
EOF
  foreach $temp (@reqqueue) {
    $urlname=uri_escape($$temp{'chan'});
    &connsend($conn,<<"EOF");
   <tr>
    <td>$$temp{'chan'}</td>
    <td>$$temp{'user'}</td>
    <td>$$temp{'type'}</td>
    <td><a href="$$temp{'page'}" target="_new">LINK</a></td>
    <td>$$temp{'ctime'}</td>
    <td>$$temp{'stats'}</td>
    <td>$$temp{'modes'}</td>
    <td>$$temp{'topic'}</td>
    <td>$$temp{'users'}</td>
    <td><a href=\"/accept/$urlname\">accept</a>/<a href=\"/reject/$urlname\">reject</a></td>
   </tr>\n
EOF
  }
  &connsend($conn,<<"EOF");
  </table>
 </body>
</html>
EOF
  $select->remove($conn);
  $conn->close;
}

# Requires http-basic-auth before triggering attempt to add Q
sub accept {
  my $conn = shift;
  my $chan = shift;
  $chan = uri_unescape($chan);

  my $oper = &requireauth($conn);
  if($oper =~ /FAILED/) {return; };

  print("\033[31mAUTH>\033[0m $oper -^\n");
  plog("AUTH $oper -^\n");

  foreach $temp (@reqqueue) {
    if($$temp{'chan'} =~ /^$chan$/) {
      push(@queue,{conn=>$conn, command=>"addchan",user=>$$temp{'user'},chan=>$$temp{'chan'},type=>$$temp{'type'},mail=>$$temp{'mail'}});
    }
  }
}

# Requires http-basic-auth before removing a channel from the queue
sub reject {
  my $conn = shift;
  my $chan = shift;
  my $worked = 0;
  $chan = uri_unescape($chan);
  my $oper = &requireauth($conn);
  if($oper =~ /FAILED/) {return; };

  print("\033[31mAUTH>\033[0m $oper -^\n");
  plog("AUTH $oper -^\n");
  $worked = &remrequest($chan);
  &connsend($conn,<<"EOF");
HTTP/1.1 200 OK
Server: $version
Connection: keep-alive
Content-Type: text/html

<html>
 <head>
  <title>Reject $chan</title>
  <style>
   body { background: black ; color: white }
  </style>
 </head>
 <body>
EOF
  if($worked) {
    &connsend($conn,"  <p>Channel Rejected</p>\n");
  } else {
    &connsend($conn,"  <p>Couldn't reject channel. May have been accepted or rejected since you last loaded queue display</p>\n");
  }
  &connsend($conn,<<"EOF");
  <p>[<a href="/queue">Back</a>]</p>
 </body>
</html>
EOF
  $select->remove($conn);
  $conn->close;
}

###############################################################################
# Utility functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Tries to send a message to IRC.
sub ircsend {
  my $message = shift;

  if(defined($irc) && $irc->opened) {
    print("\033[32m>IRC<\033[0m $message");
    $irc->send($message,0);
  } else {
    print("\033[31;1m*ERR*\033[0m IRC Connection b0rked. Falling over\n");
    &savequeue();
    die("IRC Dead... Not worth carrying on.\n");
  }
}

# Tries to send a message to a socket.
sub connsend {
  my $conn = shift;
  my $message = shift;

  if(defined($conn) && $conn->opened) {
    $conn->send($message,0);
  } else {
    # Something wrong with connection. Drop it.
    print("\033[31;1m*ERR*\033[0m Connection dropped\n");
    $select->remove($conn); # How can we drop it if it's undefined?!
    $conn->close;
    return; 
  };
}

# Adds a completed request to the request queue, @reqqueue
sub addtoqueue {
	$check=0;
	$requests++;
	print("  *** Added\n");
	plog("$$chandata{'chan'} added to request queue\n");
	push(@reqqueue,{user=>$$current{'user'},    page=>$$current{'page'},
	                mail=>$$current{'mail'},    chan=>$$chandata{'chan'},
		        ctime=>$$chandata{'ctime'}, modes=>$$chandata{'modes'},
			users=>$$chandata{'users'}, stats=>$$chandata{'stats'},
			type=>$$current{'type'},   topic=>$$chandata{'topic'}});
	&connsend($$current{'conn'},<<"EOF");
 OK.</li>
  </ul>
  <p>Channel $$chandata{'chan'} checked:</p>
  <table>
   <tr><td>Topic</td><td>$$chandata{'topic'}</td></tr>
   <tr><td>Created</td><td>$$chandata{'ctime'}</td></tr>
   <tr><td>Modes</td><td>$$chandata{'modes'}</td></tr>
   <tr><td>Users</td><td>$$chandata{'users'}</td></tr>
   <tr><td>Stats</td><td>$$chandata{'stats'}</td></tr>
  </table>
  <p>Your request has been added to the queue. If you registered your email you
   should receive a notification when Q is added.</p>
  <p>[<a href="/">Back</a>]</p>
 </body>
</html>
EOF
        $select->remove($$current{'conn'});
	$$current{'conn'}->close;
	$ready=1;
}

# Removes a request from the queue. Informs requester if possible
sub remrequest {
  my($chan)=shift;
  my(@newqueue,$worked);
  foreach my $request (@reqqueue) {
    # Search request list for channel, squish it (muahahaha)
    if($$request{'chan'} =~ /^$chan$/) {
      $worked = 1;
      $requests--;
      if($$request{'mail'} =~ /\S+\@\S+/) {
        if($$current{'command'} =~ /addchan/) {
	  # channel was accepted
	  open(MAIL, "|mutt -s \"Channel $$request{'chan'} Accepted\" $$request{'mail'}");
	  print(MAIL "Q should now be in your channel\n");
	  close(MAIL);
	} else {
	  # channel was rejected
	  open(MAIL, "|mutt -s \"Channel $$request{'chan'} Rejected\" $$request{'mail'}");
	  print(MAIL "Your channel did not fit the requirements for Q to be added\n");
	  close(MAIL);
	}
      }
    } else {
      push(@newqueue,$request);
    }
  }
  @reqqueue=@newqueue;
  $worked;
}

# Saves the current contents of the queue to a file
sub savequeue {
  print("<<<<< Saving Queue [");
  open(QUEUE, ">$queuefile");
  foreach my $temp (@reqqueue) {
    print(".");
    # typical database, \n record terminated. Surely all databases are like this
    print(QUEUE uri_escape($$temp{'user'})  . "&" .  # I have uri_escape, might
                uri_escape($$temp{'page'})  . "&" .  #   as well use it.
		uri_escape($$temp{'mail'})  . "&" . 
		uri_escape($$temp{'chan'})  . "&" . 
		uri_escape($$temp{'ctime'}) . "&" . 
		uri_escape($$temp{'modes'}) . "&" . 
		uri_escape($$temp{'users'}) . "&" . 
		uri_escape($$temp{'stats'}) . "&" . 
		uri_escape($$temp{'type'})  . "&" .
		uri_escape($$temp{'topic'}) . "\n"); 
  }
  close(QUEUE);
  print("] done.\n");
  plog("Saved Queue\n");
}

# Loads the queue from a file
sub loadqueue {
  print(">>>>> Loading Queue [");
  open(QUEUE, "$queuefile");
  while (<QUEUE>) {
    print(".");
    my($user,$page,$mail,$chan,$ctime,$modes,$users,$stats,$type,$topic) = split(/&/);
    $user = uri_unescape($user);
    $page = uri_unescape($page);
    $mail = uri_unescape($mail);
    $chan = uri_unescape($chan);
    $ctime = uri_unescape($ctime);
    $modes = uri_unescape($modes);
    $users = uri_unescape($users);
    $stats = uri_unescape($stats);
    $type = uri_unescape($type);
    $topic = uri_unescape($topic);
    push(@reqqueue,{user=>$user,   page=>$page,   mail=>$mail,   chan=>$chan,
                    ctime=>$ctime, modes=>$modes, users=>$users, stats=>$stats,
		    type=>$type,   topic=>$topic});
    $requests++;
  }
  close(QUEUE);
  print("] done.\n");
}

# Prints a message to the logfile
sub plog {
  my $message = shift;
  print(LOG "[" . gmtime() . "] $message");
}

# Parses http headers in order to see if http-basic-auth passes
sub requireauth {
  my $conn = shift;
  my $auth = "FAILED";
  my $user = "FAILED";

  while(my $line = $conn->getline) {
    if($line =~ /^Authorization: Basic (\S+)/i) { $auth = $1; };
    last if($line =~ /^\r$/);
  }
  my $userpass = decode_base64($auth);
  foreach $realuser (@users) {
    if($realuser =~ /^$userpass$/) { 
      ($user,undef) = split(':',$realuser);
      last;
    }
  }
  if($user =~ /FAILED/) {
    print("\033[31mAUTH>\033[0m Failed -^\n");
    plog(" AUTH Failed -^\n");
    &connsend($conn,<<"EOF");
HTTP/1.1 401 Authorization Required
Server: $version
Www-Authenticate: Basic realm="qweb"
Content-Type: text/html
Connection: close

<html>
 <head>
  <title>401 Authorization Required</title>
 </head>
 <body>
  <h1>401 Authorization Required</h1>
  <p>Either you're not meant to be here, you haven't got your password, or your browser is incapable of sending basic-auth.</p>
 </body>
</html>
EOF
    $select->remove($conn);
    $conn->close;
  }
  $user;
}
